unit DeskWinCNC;

(*
  INPUT STATES

  ESTOP = bit 5 SET
  XHome = bit 0 SET
  YHome = bit 1 SET
  ZHome = bit 2 SET
  AHome = bit 3 SET

*)

interface

uses
  DeskWinApi;

Const
  COMMANDFAILED = 0;
  COMMANDACCEPTED = 1;
  BUFFERFULL = 2;

  TC = 27;
  NumAxis = 4;
  Inverted = 0;
  TagAxis = 0;
  MaxEncoderCount = 0;

  JOG_SMALL_STEP: double = 0.025; // 0.25mm micro jog

var
  CNCInitialised: boolean = FALSE;
  CNCControllerFound: boolean;
  CNCComPort: integer;
  CNCDualYAxis: boolean;
  CNCVisualHome: boolean;
  InPosition: boolean;
  FirmVer: double;
  CNCEmergencyStop: boolean;
  Paused: boolean;
  Ack: Byte;
  Inputs: Byte;
  Status: integer;
  Busy: boolean;
  JogX, JogY, JogZ, JogA: double;

  MotorPosX, MotorPosY, MotorPosZ, MotorPosA: integer; // raw machine position in motor steps

  CNCXPos, CNCYPos, CNCZPos, CNCAPos: double;
  CNCHome1X, CNCHome1Y, CNCHome2X, CNCHome2Y, CNCHome1Z, CNCHome2Z, CNCHomeZ: double;
  CNCActualHomeX, CNCActualHomeY: double;

  CNCAcceleration, CNCMaxStepsPerSecond, CNCStartStepsPerSecond, CNCJogSpeed: integer;
  CNCIsHome: boolean;
  CNCScale: Byte;
  CNCMinX, CNCMaxX, CNCMinY, CNCMaxY, CNCMinZ, CNCMaxZ: double;
  CNCXStepsPerMM, CNCYStepsPerMM, CNCZStepsPerMM, CNCAStepsPerRev: double;

  CNCOutputDualYAxisDisable, CNCOutputFeederState, CNCOutputBlowState, CNCOutputVacuumState, CNCOutputDispenserState: boolean;

  DeskWinAPICommand: TDeskWinAPICommand;

function CNCInitialise: boolean;
procedure CNCHome;
function WaitUntilMoveFinished: boolean;
procedure SetOutputYaxis2Disable(State: boolean);
procedure SetOutputFeeder(State: boolean);
procedure SetOutputBlow(State: boolean);
procedure SetOutputVacuum(State: boolean);
procedure SetOutputDispenser(State: boolean);
Function SendAPICommandToController(com: TDeskWinAPICommand): Byte;
function CNCMoveHomeRelativeWithNozzleProtection(X, Y, Z, A: double; Speed: integer): boolean;
function CNCMoveHomeRelative(X, Y, Z, A: double; Speed: integer): boolean;
function CNCMove(X, Y, Z, A: double; Speed: integer): boolean;
procedure Jog;

implementation

function CNCInitialise: boolean;
begin
  CNCInitialised := FALSE;
  DeskWinClose; // ensure controller is closed

  { ****************************************************************************** }

  CNCOutputDualYAxisDisable := FALSE;
  CNCOutputFeederState := FALSE; // reset control flags
  CNCOutputBlowState := FALSE;
  CNCOutputVacuumState := FALSE;
  CNCOutputDispenserState := FALSE;

  { ****************************************************************************** }

  Status := DeskWinReset(CNCComPort, 27, 1); // open controller
  CNCControllerFound := Status = DESKWIN_2nd_OK;

  { ****************************************************************************** }
  if CNCControllerFound then
  begin
    Status := DeskWinSetMaxSPS(CNCMaxStepsPerSecond, NumAxis, Inverted, TagAxis, MaxEncoderCount);
    if Status = DESKWIN_2nd_OK then
    begin
      Status := DeskWinSetAccelleration(CNCAcceleration, CNCMaxStepsPerSecond, CNCStartStepsPerSecond, CNCScale);

      if Status = DESKWIN_2nd_OK then
      begin
        if CNCDualYAxis then
          SetOutputYaxis2Disable(FALSE); // ensure DUAL Y AXIS is ENABLED!!!
        CNCInitialised := true;
      end;
    end;
    { ****************************************************************************** }

    SetOutputYaxis2Disable(FALSE); // ensure DUAL Y AXIS is ENABLED!!!

    { ****************************************************************************** }
  end;
  Result := CNCInitialised;
end;

procedure CNCHome;
begin
  if CNCControllerFound and CNCInitialised and WaitUntilMoveFinished then
  begin
    // ******************************************************************************

    CNCActualHomeX := CNCHome1X; // reset home error
    CNCActualHomeY := CNCHome1Y;

    CNCIsHome := FALSE; // indicate homed

    // ******************************************************************************

    DeskWinAPICommand.CommandType := HOMECONTROL; // home rapidly
    DeskWinAPICommand.Dir := 0; // negative move
    DeskWinAPICommand.Axis := 12; // Z axis and A axis only
    DeskWinAPICommand.SlewSPS := 300; // slow

    Repeat
      Status := SendAPICommandToController(DeskWinAPICommand);
    until Status = COMMANDACCEPTED;

    if WaitUntilMoveFinished then
    begin
      { ****************************************************************************** }

      DeskWinAPICommand.CommandType := HOMECONTROL; // home rapidly
      DeskWinAPICommand.Dir := 0; // negative move
      DeskWinAPICommand.Axis := 15; // all axis
      DeskWinAPICommand.SlewSPS := 5000; // fast
      DeskWinAPICommand.EndSPS := 0;

      Repeat
        Status := SendAPICommandToController(DeskWinAPICommand);
      until Status = COMMANDACCEPTED;

      if WaitUntilMoveFinished then
      begin
        { ****************************************************************************** }

        DeskWinAPICommand.CommandType := SETPOS; // set absolute position
        DeskWinAPICommand.Axis := 15; // all axis
        DeskWinAPICommand.X := 0;
        DeskWinAPICommand.Y := 0;
        DeskWinAPICommand.Z := 0;
        DeskWinAPICommand.A := 0;

        Repeat
          Status := SendAPICommandToController(DeskWinAPICommand);
        until Status = COMMANDACCEPTED;

        { ****************************************************************************** }

        DeskWinAPICommand.CommandType := RAPIDFEED; // move back from home sensors
        DeskWinAPICommand.SlewSPS := 300; // slow
        DeskWinAPICommand.EndSPS := 0;
        DeskWinAPICommand.X := 400;
        DeskWinAPICommand.Y := 400;
        DeskWinAPICommand.Z := 250;
        DeskWinAPICommand.A := 0;

        Repeat
          Status := SendAPICommandToController(DeskWinAPICommand);
        until Status = COMMANDACCEPTED;

        if WaitUntilMoveFinished then
        begin
          { ****************************************************************************** }

          DeskWinAPICommand.CommandType := RAPIDFEED; // move back from home sensors
          DeskWinAPICommand.SlewSPS := 300; // slow for A axis
          DeskWinAPICommand.EndSPS := 0;
          DeskWinAPICommand.X := 400;
          DeskWinAPICommand.Y := 400;
          DeskWinAPICommand.Z := 250;
          DeskWinAPICommand.A := 400;

          Repeat
            Status := SendAPICommandToController(DeskWinAPICommand);
          until Status = COMMANDACCEPTED;

          { ****************************************************************************** }

          if WaitUntilMoveFinished then
          begin
            DeskWinAPICommand.CommandType := HOMECONTROL; // home again slowly
            DeskWinAPICommand.Dir := 0; // negative move
            DeskWinAPICommand.Axis := 15; // all axis
            DeskWinAPICommand.SlewSPS := 300; // fast
            DeskWinAPICommand.EndSPS := 0;

            Repeat
              Status := SendAPICommandToController(DeskWinAPICommand);
            until Status = COMMANDACCEPTED;

            if WaitUntilMoveFinished then
            begin
              { ****************************************************************************** }

              DeskWinAPICommand.CommandType := SETPOS; // set absolute position
              DeskWinAPICommand.Axis := 15; // all axis
              DeskWinAPICommand.X := 0;
              DeskWinAPICommand.Y := 0;
              DeskWinAPICommand.Z := 0;
              DeskWinAPICommand.A := 0;

              Repeat
                Status := SendAPICommandToController(DeskWinAPICommand);
              until (Status = COMMANDACCEPTED);

              if WaitUntilMoveFinished then
              begin
                CNCIsHome := true; // indicate homed
              end;
            end;
          end;
        end;
      end;
    end;
  end;
end;

function WaitUntilMoveFinished: boolean; // returns false if fault
begin
  if CNCControllerFound and CNCInitialised then
  begin
    Repeat
      if CNCEmergencyStop then // if emergency stop
      begin
        DeskWinAPICommand.CommandType := ENDMOVE; // end move
        Status := SendAPICommandToController(DeskWinAPICommand);
      end
      else
      begin
        InPosition := FALSE;
        DeskWinAPICommand.CommandType := GETPOS; // get position
        Status := SendAPICommandToController(DeskWinAPICommand);
      end;
    until ((Status = COMMANDACCEPTED) and InPosition) or CNCEmergencyStop;

    Result := ((Status = COMMANDACCEPTED) and InPosition) and not CNCEmergencyStop;
  end
  else
    Result := FALSE; // emergency stop
end;

procedure SetOutputYaxis2Disable(State: boolean);
begin
  if CNCControllerFound and CNCInitialised then
  begin
    DeskWinAPICommand.CommandType := OUTPUT;
    DeskWinAPICommand.OUTPUT := 4;
    if State then
      DeskWinAPICommand.State := 1
    else
      DeskWinAPICommand.State := 0;

    Repeat
      Status := SendAPICommandToController(DeskWinAPICommand);
    until (Status = COMMANDACCEPTED);

    CNCOutputDualYAxisDisable := State;
  end;
end;

procedure SetOutputFeeder(State: boolean);
begin
  if CNCControllerFound and CNCInitialised then
  begin
    DeskWinAPICommand.CommandType := OUTPUT;
    DeskWinAPICommand.OUTPUT := 8;
    if State then
      DeskWinAPICommand.State := 1
    else
      DeskWinAPICommand.State := 0;

    Repeat
      Status := SendAPICommandToController(DeskWinAPICommand);
    until (Status = COMMANDACCEPTED);

    CNCOutputFeederState := State;
  end;
end;

procedure SetOutputBlow(State: boolean);
begin
  if CNCControllerFound and CNCInitialised then
  begin
    DeskWinAPICommand.CommandType := OUTPUT;
    DeskWinAPICommand.OUTPUT := 32;
    if State then
      DeskWinAPICommand.State := 1
    else
      DeskWinAPICommand.State := 0;

    Repeat
      Status := SendAPICommandToController(DeskWinAPICommand);
    until (Status = COMMANDACCEPTED);

    CNCOutputBlowState := State;
  end;
end;

procedure SetOutputVacuum(State: boolean);
begin
  if CNCControllerFound and CNCInitialised then
  begin
    DeskWinAPICommand.CommandType := OUTPUT;
    DeskWinAPICommand.OUTPUT := 16;
    if State then
      DeskWinAPICommand.State := 1
    else
      DeskWinAPICommand.State := 0;

    Repeat
      Status := SendAPICommandToController(DeskWinAPICommand);
    until (Status = COMMANDACCEPTED);

    CNCOutputVacuumState := State;
  end;
end;

procedure SetOutputDispenser(State: boolean);
begin
  if CNCControllerFound and CNCInitialised then
  begin
    DeskWinAPICommand.CommandType := OUTPUT;
    DeskWinAPICommand.OUTPUT := 64;
    if State then
      DeskWinAPICommand.State := 1
    else
      DeskWinAPICommand.State := 0;

    Repeat
      Status := SendAPICommandToController(DeskWinAPICommand);
    until (Status = COMMANDACCEPTED);

    CNCOutputDispenserState := State;
  end;
end;

procedure Jog;
var
  X, Y, Z, A: double;
begin // jog at selected speed in selected direction
  X := CNCXPos + JogX;
  Y := CNCYPos + JogY;
  Z := CNCZPos + JogZ;
  A := CNCAPos + JogA;
  CNCMove(X, Y, Z, A, CNCJogSpeed); // move to target
end;

function CNCMoveHomeRelativeWithNozzleProtection(X, Y, Z, A: double; Speed: integer): boolean;
begin // move relative to HOME position but PROTECT NOZZLE against damage
  if (CNCZPos > 0) or (Z > 0) then
  begin // if Z is too low and X or Y movement is required
    Result := CNCMove(CNCXPos, CNCYPos, 0, A, Speed); // move Z axis home first

    if Result then
      Result := CNCMoveHomeRelative(X, Y, 0 - CNCHomeZ, A, Speed); // then move to position

    if Result then
      Result := CNCMoveHomeRelative(X, Y, Z, A, Speed); // move Z axis last
  end
  else
    Result := CNCMoveHomeRelative(X, Y, Z, A, Speed); // move directly as move is ok
end;

function CNCMoveHomeRelative(X, Y, Z, A: double; Speed: integer): boolean;
begin // move relative to HOME position
  Result := CNCMove(X + CNCActualHomeX, Y + CNCActualHomeY, Z + CNCHomeZ, A, Speed);
end;

function CNCMove(X, Y, Z, A: double; Speed: integer): boolean;
var // general move absolute
  OutOfLimits: boolean;
  X1, Y1, Z1, A1: double;
begin

  if CNCControllerFound and CNCInitialised then
  begin
    X1 := X; // COPY BEFORE APPLYING LIMITS
    Y1 := Y;
    Z1 := Z;
    A1 := A;

    if WaitUntilMoveFinished then // wait until previous move is finished
    begin
      OutOfLimits := FALSE; // assume ok

      if X1 < CNCMinX then
      begin
        X1 := CNCMinX;
        OutOfLimits := true;
      end;

      if X1 > CNCMaxX then
      begin
        X1 := CNCMaxX;
        OutOfLimits := true;
      end;

      if Y1 < CNCMinY then
      begin
        Y1 := CNCMinY;
        OutOfLimits := true;
      end;

      if Y1 > CNCMaxY then
      begin
        Y1 := CNCMaxY;
        OutOfLimits := true;
      end;

      if Z1 < CNCMinZ then
      begin
        Z1 := CNCMinZ;
        OutOfLimits := true;
      end;

      if Z1 > CNCMaxZ then
      begin
        Z1 := CNCMaxZ;
        OutOfLimits := true;
      end;

      if not CNCEmergencyStop then // if not emergency stop
      begin
        DeskWinAPICommand.CommandType := LINEARFEED; // move

        DeskWinAPICommand.SlewSPS := Speed;
        DeskWinAPICommand.EndSPS := 0;
        DeskWinAPICommand.X := Round(X1 * CNCXStepsPerMM);
        DeskWinAPICommand.Y := Round(Y1 * CNCYStepsPerMM);
        DeskWinAPICommand.Z := Round(Z1 * CNCZStepsPerMM);
        DeskWinAPICommand.A := Round(A1 * CNCAStepsPerRev / 360);

        Repeat
          Status := SendAPICommandToController(DeskWinAPICommand);
        until (Status = COMMANDACCEPTED);

        if WaitUntilMoveFinished then // wait until this move is finished
        begin
          CNCIsHome := (MotorPosX = 0) and (MotorPosY = 0) and (MotorPosZ = 0) and (MotorPosA = 0); // updated homed status

          Result := not OutOfLimits; // move ok if not out of limits
        end
        else
          Result := FALSE; // out of limits
      end
      else
        Result := FALSE; // out of limits
    end
    else
      Result := FALSE; // out of limits

    CNCXPos := X1; // update results
    CNCYPos := Y1;
    CNCZPos := Z1;
    CNCAPos := A1;

  end
  else
    Result := FALSE; // out of limits
end;

Function SendAPICommandToController(com: TDeskWinAPICommand): Byte;
var
  X, Y, Z, A: integer;
  Ack: Byte;
begin
  Result := COMMANDFAILED; // assume failure

  If Not CNCControllerFound Then
    Exit;

  Ack := 0;

  Case com.CommandType of
    ARCDIRECTION:
      Ack := DeskWinSetArcDirection(com.Dir);

    ACCELCOM:
      ; // usually sent directly

    GETVERSION:
      FirmVer := DeskWinGetFirmwareVersion();

    RESETCONTROL:
      Ack := DeskWinReset(CNCComPort, TC, 1);

    GETPOS:
      begin
        Ack := DeskWinGetPosition(MotorPosX, MotorPosY, MotorPosZ, MotorPosA);
        InPosition := Ack = DESKWIN_2nd_IN_POSITION;
      end;

    PAUSECONTROL:
      begin
        Ack := DeskWinPause;
        If Ack = DESKWIN_2nd_OK Then
          Paused := true;
      end;

    CONTINUECONTROL:
      begin
        If CNCEmergencyStop Then
          Ack := DESKWIN_2nd_IN_EMSTOP
        Else
          Ack := DeskWinContinue;
        If Ack = DESKWIN_2nd_OK Then
          Paused := FALSE;
      end;

    ENDMOVE:
      Ack := DeskWinEndCurrentMove;

    EMERGENCYSTOP:
      Ack := DeskWinEmStop;

    SETPOS:
      Ack := DeskWinSetPosition(com.X, com.Y, com.Z, com.A, com.Axis);

    OUTPUT:
      begin
        If CNCEmergencyStop Then
          Ack := DESKWIN_2nd_IN_EMSTOP
        Else
          Ack := DeskWinSetOutput(com.OUTPUT, com.State, MotorPosX, MotorPosY, MotorPosZ, MotorPosA);
      End;

    CONTROLDELAY:
      begin
        If CNCEmergencyStop Then
          Ack := DESKWIN_2nd_IN_EMSTOP
        Else
          Ack := DeskWinDelay(com.Secs, MotorPosX, MotorPosY, MotorPosZ, MotorPosA);
      End;

    LINEARFEED:
      begin
        If CNCEmergencyStop Then
          Ack := DESKWIN_2nd_IN_EMSTOP
        Else
          Ack := DeskWinLinearMove(com.X, com.Y, com.Z, com.A, com.SlewSPS, com.EndSPS, X, Y, Z, A);
      End;

    INDEXFEED:
      begin
        If CNCEmergencyStop Then
          Ack := DESKWIN_2nd_IN_EMSTOP
        Else
          Ack := DeskWinIndexMove(com.X, com.Y, com.Z, com.A, com.SlewSPS, com.Axis, X, Y, Z, A);
      End;

    RAPIDFEED:
      begin
        If CNCEmergencyStop Then
          Ack := DESKWIN_2nd_IN_EMSTOP
        Else
          Ack := DeskWinRapidMove(com.X, com.Y, com.Z, com.A, com.SlewSPS, com.EndSPS, X, Y, Z, A);
      End;

    CW:
      begin
        If CNCEmergencyStop Then
          Ack := DESKWIN_2nd_IN_EMSTOP
        Else
          Ack := DeskWinQuadrantArcMoveCW(com.rad, com.X, com.Y, com.Ex, com.Ey, com.SlewSPS, com.EndSPS, com.Quad, X, Y, Z, A);
      End;

    CCW:
      begin
        If CNCEmergencyStop Then
          Ack := DESKWIN_2nd_IN_EMSTOP
        Else
          Ack := DeskWinQuadrantArcMoveCCW(com.rad, com.X, com.Y, com.Ex, com.Ey, com.SlewSPS, com.EndSPS, com.Quad, X, Y, Z, A);
      End;

    HOMECONTROL:
      begin
        If CNCEmergencyStop Then
          Ack := DESKWIN_2nd_IN_EMSTOP
        Else
          Ack := DeskWinHome(com.Axis, com.Dir, com.SlewSPS);
      End;

    SETOVERRIDE:
      Ack := DeskWinSetFeedOverride(com.SlewSPS);

    SETOVERRIDEMAX:
      Ack := DeskWinSetMaxSPS(com.SlewSPS, NumAxis, Inverted, TagAxis, com.NumSteps);

    READINPUTS:
      begin
        Inputs := DeskWinReadInputs();
        Ack := DESKWIN_2nd_OK;
      end;

    CHASEENCODER:
      begin
        If com.Quad = 1 Then
          Ack := DeskWinChaseEncoder(com.Axis, com.Dir, com.X, com.Y, com.Z, com.A)
        Else
          Ack := DeskWinEndChaseEncoder();
      End;

    OUTPUTIMEDIATE:
      begin
        If CNCEmergencyStop Then
          Ack := DESKWIN_2nd_IN_EMSTOP
        Else
          Ack := DeskWinSetOutputImediate(com.Dir);
      End;
  End;

  Case Ack of
    0 .. 211, 214 .. 232:
      Result := COMMANDACCEPTED;

    DESKWIN_2nd_OK:
      Result := COMMANDACCEPTED;

    DESKWIN_2nd_BUFFERFULL:
      Result := BUFFERFULL;

    DESKWIN_2nd_IN_POSITION:
      begin
        InPosition := true;
        Result := COMMANDACCEPTED;
      end;

    DESKWIN_2nd_SYNTAXERROR:
      Result := COMMANDFAILED;

    DESKWIN_2nd_LIMITHIT:
      begin
        InPosition := true;
        Result := COMMANDFAILED;
      end;

    DESKWIN_2nd_INVALID_CHECKSUM:
      Result := COMMANDFAILED;

    DESKWIN_2nd_UNKNOWN_COMMAND:
      Result := COMMANDFAILED;

    DESKWIN_2nd_INVALIDSTART:
      Result := COMMANDFAILED;

    DESKWIN_2nd_PROBEHIT:
      begin
        InPosition := true;
        Result := COMMANDACCEPTED;
      end;

    DESKWIN_2nd_COMMAND_IS_IN_PROGRESS:
      Result := COMMANDFAILED;

    DESKWIN_2nd_PORT_DOES_NOT_EXIST:
      Result := COMMANDFAILED;

    DESKWIN_2nd_PORT_IS_ALREADY_OPEN:
      Result := COMMANDFAILED;

    DESKWIN_2nd_TIMEDOUT:
      Result := COMMANDFAILED;

    DESKWIN_2nd_FAILED_TO_RESET:
      Result := COMMANDFAILED;

    DESKWIN_2nd_NOTINITED:
      Result := COMMANDFAILED;
  End;
End;

begin
  CNCControllerFound := FALSE;
  CNCIsHome := FALSE;
  CNCEmergencyStop := FALSE;

End.
