unit filterBlobRepositioning;
(* ***** BEGIN LICENSE BLOCK *****
  * Copyright (C) 2004 Durand Emmanuel
  * Copyright (C) 2004 Burgel Eric
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
  * Contact :
  *   filters@edurand.com
  *   filters@burgel.com
  *
  * ***** END LICENSE BLOCK ***** *)

{
  edurand (filters@edurand.com)
  eburgel (filters@burgel.com)
}

interface

uses
  filter, fparameters, image, windows,
  filterVectorHistogram, filterBlobExplorer, filterRotation, filterCopy,
  filterResize, filterCorrelation, filterBlobBalance;

type
  TFilterBlobRepositioning = class(TFilter)
  public
    constructor Create; override;
    destructor Destroy; override;
    procedure Run(); override;
    procedure Run(command: String); override;
    procedure setParameterInteger(const aName: String; const aValue: Int64); override;
  private
    parameterImageIn: TParameterImage;
    // parameterBlob : TParameterPointer;
    parameterBlob_ThresholdBackground: TParameterInteger;
    parameterBlobAreaMin, parameterBlobAreaMax: TParameterInteger;
    parameterMonitoring: TParameterBoolean;
    parameterVectorhistogramSmooth: TParameterInteger;
    parameterVectorhistogramAngleprecision: TParameterInteger;
    parameterMargin: TParameterInteger;
    outputparameterImageOutMonitoring: TParameterImage;
    ouputParameterAngle: TParameterSingle;
    outputparameterImageOut: TParameterImage;
    inImage, outImage: PBitmap32;
    inImageBlob: PBitmap32;
    imageOutMonitorings: ArrayOfPBitmap32;
    smooth, margin: Integer;
    blobThreshold, blobAreaMin, blobAreaMax: Integer;
    blob: PBlob;
    vectorHistogram: TParameterArraySingles;
    vectorHistogramToMatch: Array of Single;
    vectorHistogramAngleprecision: Integer;
    bestAngle: Single;
    learnedImage, imageToCompareWithLearnedImage: PBitmap32;
    filterBlobExplorer: TFilterBlobExplorer;
    filterVectorHistogram: TFilterVectorHistogram;
    filterRotation, filterRotationForAccurateSearch: TFilterRotation;
    filterCopy: TFilterCopy;
    filterCorrelation: TFilterCorrelation;
    filterBlobBalance: TFilterBlobBalance;
    procedure _run();
    procedure destroyImageOuts();
    procedure createVectorHistogram();
    procedure searchAngleByVectorHistogram();
    procedure searchAccurateAngle();
    procedure extractBlobRealImage(angle: Single);
    procedure viewVectorChain(aImage: PBitmap32);
    procedure viewVectorHistogram(aImage: PBitmap32);
    procedure setDefaultVectorHistogramToMatch;
  end;

implementation

uses
  imageIO, Math, divers, SysUtils;

constructor TFilterBlobRepositioning.Create;
begin
  inherited;
  parameterImageIn := addParameterImage('INIMAGE', 'input image');
  // parameterBlob:=addParameterPointer('blob','a pointer on a TBlob object provides by TFilterBlobExplorer');
  parameterBlob_ThresholdBackground := addParameterInteger('THRESHOLDBACKGROUND', '', 0, 255, 200);
  parameterBlobAreaMin := addParameterInteger('AREAMIN', 'explication...', 0, 999999, 0);
  parameterBlobAreaMax := addParameterInteger('AREAMAX', 'explication...', 0, 999999, 999999);
  parameterMonitoring := addParameterBoolean('MONITORING', 'monitoring', True);
  parameterVectorhistogramSmooth := addParameterInteger('VECTORHISTOGRAM_SMOOTH',
    'smooth to apply on vector histogram', 0, 50, 40);
  parameterVectorhistogramAngleprecision := addParameterInteger('VECTORHISTOGRAM_ANGLEPRECISION',
    'The number of different bin to class all angle (0->360 degrees). For example, set 360 to have 1 bin by degree, or set 3600 to have 10 bins by degree.', 1, 36000, 360);
  setParameterInteger('VECTORHISTOGRAM_ANGLEPRECISION', 360);
  parameterMargin := addParameterInteger('MARGIN', 'margin to add around blob before extract it', 0, 100, 10);
  outputparameterImageOutMonitoring := addOutputParameterImage('OUTIMAGEMONITORING', 'outImages for Monitoring');
  ouputParameterAngle := addOutputParameterSingle('ANGLETORESTOREORIGINALORIENTATION', 'computed angle');
  outputparameterImageOut := addOutputParameterImage('OUTIMAGE', 'outImage');
  SetLength(imageOutMonitorings, 4);
  filterBlobExplorer := TFilterBlobExplorer.Create;
  filterVectorHistogram := TFilterVectorHistogram.Create;
  filterRotation := TFilterRotation.Create;
  filterCopy := TFilterCopy.Create;
  filterRotationForAccurateSearch := TFilterRotation.Create;
  filterCorrelation := TFilterCorrelation.Create;
  filterBlobBalance := TFilterBlobBalance.Create;
  learnedImage := nil;
  imageToCompareWithLearnedImage := nil;
  vectorHistogram := nil;
end;

destructor TFilterBlobRepositioning.Destroy;
begin
  image.freeImage(inImageBlob);
  image.freeImage(learnedImage);
  filterBlobExplorer.Free;
  filterBlobBalance.Free;
  filterCorrelation.Free;
  filterRotationForAccurateSearch.Free;
  filterCopy.Free;
  filterRotation.Free;
  filterVectorHistogram.Free;
  destroyImageOuts();
  inherited;
end;

procedure TFilterBlobRepositioning.destroyImageOuts();
var
  i: Integer;
begin
  for i := Low(imageOutMonitorings) to High(imageOutMonitorings) do
  begin
    image.freeImage(imageOutMonitorings[i]);
  end;
  // outImage is a image provided by the filterCopy filter,
  // so we do not have to delete it
end;

procedure TFilterBlobRepositioning.setParameterInteger(const aName: String; const aValue: Int64);
begin
  if aName = 'VECTORHISTOGRAM_ANGLEPRECISION' then
  begin
    if vectorHistogramAngleprecision <> aValue then
    begin
      vectorHistogramAngleprecision := aValue;
      // we set a basic vectorHistogramToMatch
      setDefaultVectorHistogramToMatch();
    end;
  end;
  inherited;
end;

procedure TFilterBlobRepositioning.setDefaultVectorHistogramToMatch;
var
  i: Integer;
begin
  SetLength(vectorHistogramToMatch, 0);
  SetLength(vectorHistogramToMatch, vectorHistogramAngleprecision);
  for i := 0 to vectorHistogramAngleprecision - 1 do
    vectorHistogramToMatch[i] := 0;
  vectorHistogramToMatch[0] := 1.2;
  vectorHistogramToMatch[vectorHistogramAngleprecision div 4] := 1;
  vectorHistogramToMatch[vectorHistogramAngleprecision div 2] := 1.1;
  vectorHistogramToMatch[3 * (vectorHistogramAngleprecision div 4)] := 1;
  divers.smoothIt(vectorHistogramToMatch, 10);
end;

procedure TFilterBlobRepositioning.Run();
begin
  inImage := parameterImageIn.image;
  if (inImage <> nil) then
  begin
    _run();
  end;
end;

procedure TFilterBlobRepositioning._run();
begin
  destroyImageOuts();

  smooth := parameterVectorhistogramSmooth.Value;
  margin := parameterMargin.Value;
  blobThreshold := parameterBlob_ThresholdBackground.Value;
  blobAreaMin := parameterBlobAreaMin.Value;
  blobAreaMax := parameterBlobAreaMax.Value;
  bestAngle := 0;

  inImageBlob := image.eraseOrCreateImageLike(inImageBlob, inImage);
  filterBlobExplorer.setParameterImage('INIMAGE', inImage);
  filterBlobExplorer.setParameterImage('OUTIMAGE', inImageBlob);
  filterBlobExplorer.setParameterInteger('INTENSITYBACKGROUND', blobThreshold);
  filterBlobExplorer.setParameterInteger('INTENSITYPRECISION', 40);
  filterBlobExplorer.setParameterBoolean('ENABLEAREA', TRUE);
  filterBlobExplorer.setParameterInteger('AREAMIN', blobAreaMin);
  filterBlobExplorer.setParameterInteger('AREAMAX', blobAreaMax);
  filterBlobExplorer.setParameterBoolean('CONTOUR', false);
  filterBlobExplorer.setParameterBoolean('CRITICALPOINTS', True);
  filterBlobExplorer.setParameterInteger('APPROXIMATIONACCURACY', 20);
  filterBlobExplorer.setParameterString('APPROXIMATIONMETHOD', 'Douglas-Peucker');
  filterBlobExplorer.setParameterBoolean('SURFACEINFO', True);
  filterBlobExplorer.setParameterBoolean('MONITORING', false);
  filterBlobExplorer.Run;
  if (Length(filterBlobExplorer.getOutputParameterArrayPointers('BLOBS').Pointers) > 0) then
  begin
    blob := filterBlobExplorer.getOutputParameterArrayPointers('BLOBS').Pointers[0];
    searchAngleByVectorHistogram();
    searchAccurateAngle();
    if parameterMonitoring.Value = True then
    begin
      // show approximatedSegmentList
      imageOutMonitorings[0] := image.createImageFromImage(inImage);
      image.drawLines(imageOutMonitorings[0], blob.approximatedSegmentList, clYellow32, clRed32);
      // show vector chain
      imageOutMonitorings[1] := image.createImage(Floor(blob.perimeter) + 1, 360);
      viewVectorChain(imageOutMonitorings[1]);
      // show vector histogram
      imageOutMonitorings[2] := image.createImage(vectorHistogramAngleprecision, 200);
      viewVectorHistogram(imageOutMonitorings[2]);
      // learned image
      if imageOutMonitorings[3] = nil then
      begin
        if learnedImage <> nil then
        begin
          imageOutMonitorings[3] := image.createImageFromImage(learnedImage);
        end
        else
        begin
          imageOutMonitorings[3] := image.createImage(0, 0);
        end;
      end;
      // set output parameter
      setOutputParameterImages('OUTIMAGEMONITORING', imageOutMonitorings);
    end;
  end
  else
  begin
    outImage := image.createImage(1, 1);
  end;
  setOutputParameterSingle('ANGLEOUTPUTORIENTATION', bestAngle);
  setOutputParameterImage('OUTIMAGE', outImage);
end;

procedure TFilterBlobRepositioning.Run(command: String);
var
  i: Integer;
begin
  if (command = 'LEARN') then
  begin
    image.freeImage(learnedImage);
    Run;
    if vectorHistogram <> nil then
    begin
      for i := 0 to vectorHistogramAngleprecision - 1 do
      begin
        vectorHistogramToMatch[i] := vectorHistogram.Singles[i];
      end;
    end;
    Run;
    learnedImage := image.createImageFromImage(outImage);
  end
  else
    inherited Run(command);
end;

procedure TFilterBlobRepositioning.createVectorHistogram();
begin
  filterVectorHistogram.setParameterPointer('VECTORARRAY', blob.vectorChain);
  filterVectorHistogram.setParameterInteger('ANGLEPRECISION', vectorHistogramAngleprecision);
  filterVectorHistogram.setParameterInteger('SMOOTH', smooth);
  filterVectorHistogram.Run;
  vectorHistogram := filterVectorHistogram.getOutputParameterArraySingles('VECTORHISTOGRAM');
end;

procedure TFilterBlobRepositioning.searchAngleByVectorHistogram();
var
  x: Integer;
  index: Integer;
  angle: Integer;
  sum, maxSum: Extended;
  factor: Single;
begin
  bestAngle := 0;
  createVectorHistogram();
  maxSum := 0;
  factor := vectorHistogramAngleprecision / 360;
  for angle := 0 to vectorHistogramAngleprecision - 1 do
  begin
    sum := 0;
    for x := 0 to vectorHistogramAngleprecision - 1 do
    begin
      index := x + angle;
      if index >= vectorHistogramAngleprecision then
        dec(index, vectorHistogramAngleprecision);
      sum := sum + vectorHistogram.Singles[x] * vectorHistogramToMatch[index];
    end;
    if sum > maxSum then
    begin
      bestAngle := -angle / factor;
      maxSum := sum;
    end;
  end;
end;

procedure TFilterBlobRepositioning.extractBlobRealImage(angle: Single);
var
  outImageRotation: PBitmap32;
  roi: TRect;
  minX, maxX, minY, maxY: Integer;
  center: TFPoint;
  i, iMax: Integer;
  bestAngleR: Single;
  deltaWdiv2, deltaHdiv2: Integer;

  procedure searchMinMaxXY(inPoint: TFPoint);
  var
    outPoint: TFPoint;
  begin
    outPoint := divers.getRotatedPoint(inPoint, center, -bestAngleR);
    outPoint.x := outPoint.x + deltaWdiv2;
    outPoint.y := outPoint.y + deltaHdiv2;
    minX := Min(minX, Floor(outPoint.x));
    maxX := Max(maxX, Floor(outPoint.x));
    minY := Min(minY, Floor(outPoint.y));
    maxY := Max(maxY, Floor(outPoint.y));
  end;

begin
  // We start the repositioning by restoring the angle of the image
  filterRotation.setParameterImage('INIMAGE', inImage);
  filterRotation.setParameterBoolean('MONITORING', false);
  filterRotation.setParameterSingle('ANGLE', angle);
  filterRotation.setParameterInteger('INTERPOLATION', 2);
  filterRotation.setParameterString('MISSINGPIXELCOLORMODE', 'BLACK');
  filterRotation.setParameterBoolean('AUTOADJUSTSIZE', True);
  filterRotation.Run;
  outImageRotation := filterRotation.getOutputParameterImage('OUTIMAGE').image;
  // Next, we extract only the rectangle containing the blob.
  // To do this, we rotate all pixel of the contour of the blob, and then
  // search the rectangle containing this rotated pixel
  // But the image after the rotation may be bigger than source image,
  // so we need to calcul this difference in this pixel rotation
  deltaWdiv2 := (outImageRotation.Width - inImage.Width) div 2;
  deltaHdiv2 := (outImageRotation.Height - inImage.Height) div 2;
  center := fpoint(inImage.Width div 2, inImage.Height div 2);
  bestAngleR := DegToRad(angle);
  minX := MAXINT;
  maxX := -MAXINT;
  minY := MAXINT;
  maxY := -MAXINT;
  iMax := Length(blob.approximatedSegmentList) - 1;
  for i := 0 to iMax do
  begin
    searchMinMaxXY(blob.approximatedSegmentList[i].p1);
  end;
  roi.Left := minX - margin;
  roi.Top := minY - margin;
  roi.Right := maxX + margin;
  roi.Bottom := maxY + margin;
  filterCopy.setParameterImage('INIMAGE', outImageRotation);
  filterCopy.setRegionOfInterest(roi);
  filterCopy.Run;
  // then we have our repositioned blob
  outImage := filterCopy.getOutputParameterImage('OUTIMAGE').image;
end;

procedure TFilterBlobRepositioning.viewVectorHistogram(aImage: PBitmap32);
Var
  x, y: Integer;
Begin
  image.eraseImage(aImage);
  for x := 0 to vectorHistogramAngleprecision - 1 do
  Begin
    y := round(vectorHistogram.Singles[x]);
    if y >= 200 then
      y := 199;
    image.drawLine(aImage, x, 199, x, 199 - y, clRed32);
  end;
End;

procedure TFilterBlobRepositioning.viewVectorChain(aImage: PBitmap32);
var
  i, iMax: Integer;
  x, j: Integer;
  h2: Array of Single;
begin
  SetLength(h2, aImage.Width);
  x := 0;
  iMax := Length(blob.vectorChain) - 1;
  for i := 0 to iMax do
  begin
    for j := 0 to Floor(blob.vectorChain[i].Length) - 1 do
    begin
      h2[x] := RadToDeg(blob.vectorChain[i].angle);
      Inc(x);
    end;
  end;
  smoothIt(h2, smooth);
  iMax := Length(h2) - 1;
  for i := 0 to iMax do
  begin
    image.drawLine(aImage, i, aImage.Height - 1, i, aImage.Height - 1 - h2[i], clRed32);
  end;
  image.drawLine(aImage, 0, aImage.Height - 1, aImage.Width - 1, aImage.Height - 1, clGray32, 2);
  SetLength(h2, 0);
end;

procedure TFilterBlobRepositioning.searchAccurateAngle();
var
  angleDelta, angleDeltaMin, angleDeltaMax, angleDeltaInc: Single;
  bestCorrelation, correlationForAngleDelta: Single;
  accurateAngle: Single;

  procedure calculCorrelationForAngleDelta();
  begin
    filterRotationForAccurateSearch.setParameterImage('INIMAGE', outImage);
    filterRotationForAccurateSearch.setParameterBoolean('MONITORING', false);
    filterRotationForAccurateSearch.setParameterSingle('ANGLE', angleDelta);
    filterRotationForAccurateSearch.setParameterInteger('INTERPOLATION', 2);
    filterRotationForAccurateSearch.setParameterString('MISSINGPIXELCOLORMODE', 'MIRROR');
    filterRotationForAccurateSearch.setParameterBoolean('AUTOADJUSTSIZE', false);
    filterRotationForAccurateSearch.setParameterSingle('XCENTER', outImage.Width div 2);
    filterRotationForAccurateSearch.setParameterSingle('YCENTER', outImage.Height div 2);
    filterRotationForAccurateSearch.Run;
    imageToCompareWithLearnedImage := filterRotationForAccurateSearch.getOutputParameterImage('OUTIMAGE').image;
    filterCorrelation.setParameterImage('INIMAGE1', learnedImage);
    filterCorrelation.setParameterImage('INIMAGE2', imageToCompareWithLearnedImage);
    filterCorrelation.Run;
    correlationForAngleDelta := filterCorrelation.getOutputParameterSingle('CORRELATION').Value;
  end;

begin
  if learnedImage <> nil then
  begin
    // we extract the image of the blob with the current best angle
    extractBlobRealImage(bestAngle);
    // first, we search if there is an error of 180
    angleDelta := 0;
    calculCorrelationForAngleDelta;
    bestCorrelation := correlationForAngleDelta;
    angleDelta := 180;
    calculCorrelationForAngleDelta;
    // if the correlation at 180 is better than correlation at 0
    if bestCorrelation < correlationForAngleDelta then
    begin
      // then we have to correct the angle of 180
      bestAngle := bestAngle + 180;
      extractBlobRealImage(bestAngle);
    end;
    // now we search the best accurate angle
    accurateAngle := bestAngle;
    bestCorrelation := 0;
    angleDeltaInc := 0.1;
    angleDeltaMin := -1;
    angleDeltaMax := +1;
    angleDelta := angleDeltaMin;
    repeat
      calculCorrelationForAngleDelta();
      // if the correlation is better, then the angle is better
      if correlationForAngleDelta > bestCorrelation then
      begin
        bestCorrelation := correlationForAngleDelta;
        accurateAngle := bestAngle + angleDelta;
      end;
      angleDelta := angleDelta + angleDeltaInc;
    until angleDelta > angleDeltaMax;
    bestAngle := accurateAngle;
  end;
  // extract the blob with the best found angle
  extractBlobRealImage(bestAngle);
end;

end.
