unit Settings;

interface

Type
  PartRec = record
    Designator: string;
    Comment: string;
    X: double;
    Y: double;
    Angle: double;
    Feeder: integer;
    Enabled: boolean;
  end;

Type
  FeederRec = record
    Kind: integer; // 0 = reel 1 = tray 2 = vibratory
    Location: integer; // 0 = front 1 = rear 2 = right
    Description: string;
    SnapName: string;
    X: double;
    Y: double;
    Z: double; // top of part to pick
    PartHeight: double; // part height for calc of board place height
    Nozzle: integer;
    TraySpacingX: double;
    TraySpacingY: double;
    TrayCountX: integer;
    TrayCountY: integer;
    TrayPickCount: integer;
    PlaceSpeedType: integer; // fast, medium, slow
    RejectTray: integer;
    Use_Vision: boolean;
  end;

const
  Project_Extension = '.PickAndPlace';
  Feeder_Extension = '.Feeder';

var
  Data_Directory: string = '\Pick And Place Projects\';
  Settings_Directory: string = '\Pick And Place System Settings\';

  LastProjectName: String = '';
  LastProjectPath: string = '';
  ProjectName: String = '';
  ProjectPath: string = '';
  ProjectOpen: boolean = false;
  ProjectEdited: boolean = false;
  ViewSelection: integer = 0; // default to downlooking camera  0  = Downlooking, 1 = Uplooking

  TrayRefillRequest: integer = 0; // tray number needing refilling 0 = none

  AutomaticPanel: boolean = false; // flag for automatic panel cycle
  AutomaticBoard: boolean = false; // flag for automatic board cycle
  Pause: boolean = false; // flag for pause of automatic cycle

  FiducialX: double = 0; // offset error
  FiducialY: double = 0;
  FiducialXScale: double = 1; // scaling error
  FiducialYScale: double = 1;
  FiducialAngle: double = 0; // not used at present

  Tfile: textfile; // project file

  PartCount: integer;
  PartList: array of PartRec;
  CurrentPart: integer;

  PanelOriginX, PanelOriginY, PanelSpacingX, PanelSpacingY: double;
  PanelCountX, PanelCountY, PanelBoardCount: integer;

  BoardOriginX, BoardOriginY, BoardHeight: double;
  BoardFiducial1SnapName, BoardFiducial2SnapName: string;
  BoardFiducial1X, BoardFiducial1Y, BoardFiducial1Z, BoardFiducial2X, BoardFiducial2Y, BoardFiducial2Z: double;
  BoardFiducialEnable, BoardDualFiducial: boolean;
  BoardMetric: boolean;

  RejectTray1X, RejectTray2X, RejectTray3X, RejectTray4X, RejectTray5X: double;
  RejectTray1Y, RejectTray2Y, RejectTray3Y, RejectTray4Y, RejectTray5Y: double;

  FeederCount: integer;
  FeederList: array of FeederRec;
  CurrentFeeder: integer;

  ParkX, ParkY: double; // park position

  PickDelay, PlaceDelay, ActuatorOnDelay, ActuatorOffDelay, BlowDelay: double;
  FastPlaceSpeed, MediumPlaceSpeed, SlowPlaceSpeed: integer;
  PickRetryCount: integer;
  PlaceZPush: double; // extra distance to ensure firm placement

  UplookingViewWidthInMM, DownlookingViewWidthInMM: double;
  UplookingCameraX, UplookingCameraY, UplookingCameraZ: double;
  UplookingCameraSettlingDelay, DownlookingCameraSettlingDelay: double;
  UplookingCameraEnlargementMM, DownlookingCameraEnlargementMM: integer;
  UplookingCameraAngleTestLimit: integer;
  VisionTestMode: boolean;
  VisionCorrelationLimit: double;

  NozzleCalibrateOffsetX, NozzleCalibrateOffsetY, NozzleCalibrateOffsetZ: double;
  NozzleOffsetX, NozzleOffsetY: double;
  NozzleOffsetNozzle: integer;

  NozzleHorizontal, NozzleUp, NozzleDown, NozzlePickDown: double;
  Nozzle1X, Nozzle2X, Nozzle3X, Nozzle4X, Nozzle5X, Nozzle6X, Nozzle7X, Nozzle8X: double;
  Nozzle1Y, Nozzle2Y, Nozzle3Y, Nozzle4Y, Nozzle5Y, Nozzle6Y, Nozzle7Y, Nozzle8Y: double;
  NozzleChangerEnabled: boolean;
  NozzleChangerLocation: integer; // 0 = right, 1 = front, 2 = rear
  NozzleChangerSpeed: integer;
  CurrentNozzle: integer;

  DisplayMetric: boolean;

  PartPicked: boolean;
  PartPickedX, PartPickedY, PartPickedAngle: double; // corrected part information

procedure Load_Settings; { load all settings }
Procedure Save_Settings; { save all settings }
Procedure Save_Status; { save status related settings }
function Load_Project(Path, Name: string): boolean; { load project }
function Save_Project(Path, Name: string): boolean; { save project }
function Create_New_Project(Path, Name: string): boolean; { create project }
function Load_Feeder(Path, Name: string; var Feeder: FeederRec): boolean; { load feeder }
function Save_Feeder(Path, Name: string; const Feeder: FeederRec): boolean; { save feeder }

implementation

uses Windows, IniFiles, ShFolder, Registry, SysUtils, VisionSettings, DeskWinCNC;

var
  FRegIni: TRegIniFile;
  FIni: TMemIniFile;

Function ReadRegString(Section, Entry, Default: String): String;
begin
  Result := FRegIni.ReadString(Section, Entry, Default);
end;

Function ReadRegInteger(Section, Entry: String; Default: longint): longint;
begin
  Result := FRegIni.ReadInteger(Section, Entry, Default);
end;

Function ReadRegBoolean(Section, Entry: String; Default: boolean): boolean;
begin
  Result := FRegIni.ReadBool(Section, Entry, Default);
end;

Function ReadRegFloat(Section, Entry: String; Default: double): double;
var
  S: String;
begin
  S := FRegIni.ReadString(Section, Entry, '');
  if S <> '' then
  begin
    Try
      Result := StrToFloat(S);
    except
      Result := Default;
    end;
  end
  else
    Result := Default;
end;

procedure WriteRegString(Section, Entry, Str: String);
begin
  FRegIni.WriteString(Section, Entry, Str);
end;

procedure WriteRegInteger(Section, Entry: String; Int: longint);
begin
  FRegIni.WriteInteger(Section, Entry, Int);
end;

procedure WriteRegFloat(Section, Entry: String; R: double);
Var
  Str: String;
begin
  Str := FloatToStrF(R, ffFixed, 10, 10);
  WriteRegString(Section, Entry, Str);
end;

procedure WriteRegBoolean(Section, Entry: String; B: boolean);
begin
  FRegIni.WriteBool(Section, Entry, B);
end;

procedure Load_Settings; { load all settings }
begin
  FRegIni := TRegIniFile.Create('Software\Dockerty Electronics\Pick And Place 2');
  try
    CNCComPort := ReadRegInteger('CNC Settings', 'COM Port', 1);
    CNCDualYAxis := ReadRegBoolean('CNC Settings', 'Dual Y Axis', True);
    CNCVisualHome := ReadRegBoolean('CNC Settings', 'Visual Home', True);

    CNCAcceleration := ReadRegInteger('CNC Settings', 'Acceleration', 3000000);
    CNCMaxStepsPerSecond := ReadRegInteger('CNC Settings', 'Max Steps Per Second', 65000);
    CNCStartStepsPerSecond := ReadRegInteger('CNC Settings', 'Start Steps Per Second', 750);
    CNCScale := ReadRegInteger('CNC Settings', 'Scale', 64);

    CNCJogSpeed := ReadRegInteger('CNC Settings', 'Jog Speed', 3000);

    CNCMinX := ReadRegFloat('CNC Settings', 'Minimum X Position', -5);
    CNCMaxX := ReadRegFloat('CNC Settings', 'Maximum X Position', 405);

    CNCMinY := ReadRegFloat('CNC Settings', 'Minimum Y Position', -5);
    CNCMaxY := ReadRegFloat('CNC Settings', 'Maximum Y Position', 350);

    CNCMinZ := ReadRegFloat('CNC Settings', 'Minimum Z Position', 0);
    CNCMaxZ := ReadRegFloat('CNC Settings', 'Maximum Z Position', 35);

    CNCXStepsPerMM := ReadRegFloat('CNC Settings', 'X Steps Per MM', 78.74015748);
    CNCYStepsPerMM := ReadRegFloat('CNC Settings', 'Y Steps Per MM', 78.74015748);
    CNCZStepsPerMM := ReadRegFloat('CNC Settings', 'Z Steps Per MM', 78.74015748);
    CNCAStepsPerRev := ReadRegFloat('CNC Settings', 'A Steps Per Rev', 8000);

    CNCHome1X := ReadRegFloat('Home Settings', 'Home 1 X', 20);
    CNCHome1Y := ReadRegFloat('Home Settings', 'Home 1 Y', 20);
    CNCHome1Z := ReadRegFloat('Home Settings', 'Home 1 Z', 0);
    CNCHome2X := ReadRegFloat('Home Settings', 'Home 2 X', 20);
    CNCHome2Y := ReadRegFloat('Home Settings', 'Home 2 Y', 120);
    CNCHome2Z := ReadRegFloat('Home Settings', 'Home 2 Z', 0);
    CNCHomeZ := ReadRegFloat('CNC Settings', 'Home Z', 0);

    ParkX := ReadRegFloat('Home Settings', 'Park X', 0);
    ParkY := ReadRegFloat('Home Settings', 'Park Y', 0);

    PlaceZPush := ReadRegFloat('Part Handling Settings', 'Place Z Push', 0.5);
    PickDelay := ReadRegFloat('Part Handling Settings', 'Pick Delay', 0.1);
    PlaceDelay := ReadRegFloat('Part Handling Settings', 'Place Delay', 0.1);
    ActuatorOnDelay := ReadRegFloat('Part Handling Settings', 'Actuator On Delay', 0.25);
    ActuatorOffDelay := ReadRegFloat('Part Handling Settings', 'Actuator Off Delay', 0.25);
    BlowDelay := ReadRegFloat('Part Handling Settings', 'Blow Delay', 0);

    FastPlaceSpeed := ReadRegInteger('Part Handling Settings', 'Fast Place Speed', 35000);
    MediumPlaceSpeed := ReadRegInteger('Part Handling Settings', 'Medium Place Speed', 18000);
    SlowPlaceSpeed := ReadRegInteger('Part Handling Settings', 'Slow Place Speed', 8000);

    RejectTray1X := ReadRegFloat('Part Handling Settings', 'Reject Tray 1 X', 0);
    RejectTray1Y := ReadRegFloat('Part Handling Settings', 'Reject Tray 1 Y', 0);
    RejectTray2X := ReadRegFloat('Part Handling Settings', 'Reject Tray 2 X', 0);
    RejectTray2Y := ReadRegFloat('Part Handling Settings', 'Reject Tray 2 Y', 0);
    RejectTray3X := ReadRegFloat('Part Handling Settings', 'Reject Tray 3 X', 0);
    RejectTray3Y := ReadRegFloat('Part Handling Settings', 'Reject Tray 3 Y', 0);
    RejectTray4X := ReadRegFloat('Part Handling Settings', 'Reject Tray 4 X', 0);
    RejectTray4Y := ReadRegFloat('Part Handling Settings', 'Reject Tray 4 Y', 0);
    RejectTray5X := ReadRegFloat('Part Handling Settings', 'Reject Tray 1 X', 0);
    RejectTray5Y := ReadRegFloat('Part Handling Settings', 'Reject Tray 1 Y', 0);

    PickRetryCount := ReadRegInteger('Part Handling Settings', 'Pick Retry Count', 2);

    FormVisionSettings.CameraDeviceNameUp := ReadRegString('Camera Settings', 'Camera Up Device Name', '');
    FormVisionSettings.CameraFormatNameUp := ReadRegString('Camera Settings', 'Camera Up Format Name', '');
    FormVisionSettings.CameraDeviceNameDown := ReadRegString('Camera Settings', 'Camera Down Device Name', '');
    FormVisionSettings.CameraFormatNameDown := ReadRegString('Camera Settings', 'Camera Down Format Name', '');

    UplookingViewWidthInMM := ReadRegFloat('Camera Settings', 'Uplooking Cameara View Width in MM', 40);
    UplookingCameraAngleTestLimit := ReadRegInteger('Camera Settings', 'Uplooking Camera Angle Test Limit', 40);
    UplookingCameraEnlargementMM := ReadRegInteger('Camera Settings', 'Uplooking Camera Enlargement MM', 6);
    UplookingCameraX := ReadRegFloat('Camera Settings', 'Uplooking Camera X', 200);
    UplookingCameraY := ReadRegFloat('Camera Settings', 'Uplooking Camera Y', 200);
    UplookingCameraZ := ReadRegFloat('Camera Settings', 'Uplooking Camera Z', 0);
    UplookingCameraSettlingDelay := ReadRegFloat('Camera Settings', 'Uplooking Camera Settling Delay', 0.25);
    DownlookingViewWidthInMM := ReadRegFloat('Camera Settings', 'Downlooking Cameara View Width in MM', 40);
    DownlookingCameraSettlingDelay := ReadRegFloat('Camera Settings', 'Downlooking Camera Settling Delay', 0.25);
    DownlookingCameraEnlargementMM := ReadRegInteger('Camera Settings', 'Downlooking Camera Enlargement MM', 6);
    VisionCorrelationLimit := ReadRegFloat('Camera Settings', 'Vision Correlation Limit', 0.7);

    NozzleOffsetX := ReadRegFloat('Nozzle Offset Settings', 'Nozzle Offset X', 0);
    NozzleOffsetY := ReadRegFloat('Nozzle Offset Settings', 'Nozzle Offset Y', 0);
    NozzleOffsetNozzle := ReadRegInteger('Nozzle Offset Settings', 'Nozzle Number', 1);

    NozzleCalibrateOffsetX := ReadRegFloat('Nozzle Offset Settings', 'Nozzle Calibration X', 0);
    NozzleCalibrateOffsetY := ReadRegFloat('Nozzle Offset Settings', 'Nozzle Calibration Y', 0);
    NozzleCalibrateOffsetZ := ReadRegFloat('Nozzle Offset Settings', 'Nozzle Calibration Z', 0);

    NozzleHorizontal := ReadRegFloat('Nozzle Changer Settings', 'Nozzle Horizontal', 16);
    NozzleUp := ReadRegFloat('Nozzle Changer Settings', 'Nozzle Up', 10);
    NozzleDown := ReadRegFloat('Nozzle Changer Settings', 'Nozzle Down', 20);
    NozzlePickDown := ReadRegFloat('Nozzle Changer Settings', 'Nozzle Pick Down', 0);
    NozzleChangerSpeed := ReadRegInteger('Nozzle Changer Settings', 'Nozzle Changer Speed', 3000);

    Nozzle1X := ReadRegFloat('Nozzle Changer Settings', 'Nozzle 1 X', 0);
    Nozzle2X := ReadRegFloat('Nozzle Changer Settings', 'Nozzle 2 X', 0);
    Nozzle3X := ReadRegFloat('Nozzle Changer Settings', 'Nozzle 3 X', 0);
    Nozzle4X := ReadRegFloat('Nozzle Changer Settings', 'Nozzle 4 X', 0);
    Nozzle5X := ReadRegFloat('Nozzle Changer Settings', 'Nozzle 5 X', 0);
    Nozzle6X := ReadRegFloat('Nozzle Changer Settings', 'Nozzle 6 X', 0);
    Nozzle7X := ReadRegFloat('Nozzle Changer Settings', 'Nozzle 7 X', 0);
    Nozzle8X := ReadRegFloat('Nozzle Changer Settings', 'Nozzle 8 X', 0);
    Nozzle1Y := ReadRegFloat('Nozzle Changer Settings', 'Nozzle 1 Y', 0);
    Nozzle2Y := ReadRegFloat('Nozzle Changer Settings', 'Nozzle 2 Y', 0);
    Nozzle3Y := ReadRegFloat('Nozzle Changer Settings', 'Nozzle 3 Y', 0);
    Nozzle4Y := ReadRegFloat('Nozzle Changer Settings', 'Nozzle 4 Y', 0);
    Nozzle5Y := ReadRegFloat('Nozzle Changer Settings', 'Nozzle 5 Y', 0);
    Nozzle6Y := ReadRegFloat('Nozzle Changer Settings', 'Nozzle 6 Y', 0);
    Nozzle7Y := ReadRegFloat('Nozzle Changer Settings', 'Nozzle 7 Y', 0);
    Nozzle8Y := ReadRegFloat('Nozzle Changer Settings', 'Nozzle 8 Y', 0);

    NozzleChangerEnabled := ReadRegBoolean('Nozzle Changer Settings', 'Nozzle Changer Enabled', True);
    NozzleChangerLocation := ReadRegInteger('Nozzle Changer Settings', 'Nozzle Changer Location', 0);

    CurrentNozzle := ReadRegInteger('Current Status', 'Current Nozzle', 0);
    VisionTestMode := ReadRegBoolean('Current Status', 'Vision Test Mode', false);

    LastProjectPath := ReadRegString('Current Status', 'Last Project Path', '');
    LastProjectName := ReadRegString('Current Status', 'Last Project Name', '');

  finally
  end;
  FRegIni.Free;
end;

Procedure Save_Settings; { save all settings }
begin
  FRegIni := TRegIniFile.Create('Software\Dockerty Electronics\Pick And Place 2');
  try
    WriteRegInteger('CNC Settings', 'COM Port', CNCComPort);

    WriteRegBoolean('CNC Settings', 'Dual Y Axis', CNCDualYAxis);
    WriteRegBoolean('CNC Settings', 'Visual Home', CNCVisualHome);

    WriteRegInteger('CNC Settings', 'Acceleration', CNCAcceleration);
    WriteRegInteger('CNC Settings', 'Max Steps Per Second', CNCMaxStepsPerSecond);
    WriteRegInteger('CNC Settings', 'Start Steps Per Second', CNCStartStepsPerSecond);
    WriteRegInteger('CNC Settings', 'Scale', CNCScale);

    WriteRegInteger('CNC Settings', 'Jog Speed', CNCJogSpeed);

    WriteRegFloat('CNC Settings', 'Minimum X Position', CNCMinX);
    WriteRegFloat('CNC Settings', 'Maximum X Position', CNCMaxX);
    WriteRegFloat('CNC Settings', 'Minimum Y Position', CNCMinY);
    WriteRegFloat('CNC Settings', 'Maximum Y Position', CNCMaxY);
    WriteRegFloat('CNC Settings', 'Minimum Z Position', CNCMinZ);
    WriteRegFloat('CNC Settings', 'Maximum Z Position', CNCMaxZ);
    WriteRegFloat('CNC Settings', 'X Steps Per MM', CNCXStepsPerMM);
    WriteRegFloat('CNC Settings', 'Y Steps Per MM', CNCYStepsPerMM);
    WriteRegFloat('CNC Settings', 'Z Steps Per MM', CNCZStepsPerMM);
    WriteRegFloat('CNC Settings', 'A Steps Per Rev', CNCAStepsPerRev);

    WriteRegFloat('Home Settings', 'Park X', ParkX);
    WriteRegFloat('Home Settings', 'Park Y', ParkY);

    WriteRegFloat('Home Settings', 'Home 1 X', CNCHome1X);
    WriteRegFloat('Home Settings', 'Home 1 Y', CNCHome1Y);
    WriteRegFloat('Home Settings', 'Home 1 Z', CNCHome1Z);
    WriteRegFloat('Home Settings', 'Home 2 X', CNCHome2X);
    WriteRegFloat('Home Settings', 'Home 2 Y', CNCHome2Y);
    WriteRegFloat('Home Settings', 'Home 2 Z', CNCHome2Z);
    WriteRegFloat('CNC Settings', 'Home Z', CNCHomeZ);

    WriteRegFloat('Part Handling Settings', 'Place Z Push', PlaceZPush);
    WriteRegFloat('Part Handling Settings', 'Pick Delay', PickDelay);
    WriteRegFloat('Part Handling Settings', 'Place Delay', PlaceDelay);
    WriteRegFloat('Part Handling Settings', 'Actuator On Delay', ActuatorOnDelay);
    WriteRegFloat('Part Handling Settings', 'Actuator Off Delay', ActuatorOffDelay);
    WriteRegFloat('Part Handling Settings', 'Blow Delay', BlowDelay);

    WriteRegInteger('Part Handling Settings', 'Fast Place Speed', FastPlaceSpeed);
    WriteRegInteger('Part Handling Settings', 'Medium Place Speed', MediumPlaceSpeed);
    WriteRegInteger('Part Handling Settings', 'Slow Place Speed', SlowPlaceSpeed);

    WriteRegFloat('Part Handling Settings', 'Reject Tray 1 X', RejectTray1X);
    WriteRegFloat('Part Handling Settings', 'Reject Tray 1 Y', RejectTray1Y);
    WriteRegFloat('Part Handling Settings', 'Reject Tray 2 X', RejectTray2X);
    WriteRegFloat('Part Handling Settings', 'Reject Tray 2 Y', RejectTray2Y);
    WriteRegFloat('Part Handling Settings', 'Reject Tray 3 X', RejectTray3X);
    WriteRegFloat('Part Handling Settings', 'Reject Tray 3 Y', RejectTray3Y);
    WriteRegFloat('Part Handling Settings', 'Reject Tray 4 X', RejectTray4X);
    WriteRegFloat('Part Handling Settings', 'Reject Tray 4 Y', RejectTray4Y);
    WriteRegFloat('Part Handling Settings', 'Reject Tray 5 X', RejectTray5X);
    WriteRegFloat('Part Handling Settings', 'Reject Tray 5 Y', RejectTray5Y);

    WriteRegInteger('Part Handling Settings', 'Pick Retry Count', PickRetryCount);

    WriteRegString('Camera Settings', 'Camera Up Device Name', FormVisionSettings.CameraDeviceNameUp);
    WriteRegString('Camera Settings', 'Camera Up Format Name', FormVisionSettings.CameraFormatNameUp);
    WriteRegString('Camera Settings', 'Camera Down Device Name', FormVisionSettings.CameraDeviceNameDown);
    WriteRegString('Camera Settings', 'Camera Down Format Name', FormVisionSettings.CameraFormatNameDown);

    WriteRegFloat('Camera Settings', 'Uplooking Cameara View Width in MM', UplookingViewWidthInMM);
    WriteRegFloat('Camera Settings', 'Downlooking Cameara View Width in MM', DownlookingViewWidthInMM);
    WriteRegFloat('Camera Settings', 'Uplooking Camera X', UplookingCameraX);
    WriteRegFloat('Camera Settings', 'Uplooking Camera Y', UplookingCameraY);
    WriteRegFloat('Camera Settings', 'Uplooking Camera Z', UplookingCameraZ);
    WriteRegFloat('Camera Settings', 'Uplooking Camera Settling Delay', UplookingCameraSettlingDelay);
    WriteRegInteger('Camera Settings', 'Uplooking Camera Angle Test Limit', UplookingCameraAngleTestLimit);
    WriteRegInteger('Camera Settings', 'Uplooking Camera Enlargement MM', UplookingCameraEnlargementMM);
    WriteRegFloat('Camera Settings', 'Downlooking Camera Settling Delay', DownlookingCameraSettlingDelay);
    WriteRegInteger('Camera Settings', 'Downlooking Camera Enlargement MM', DownlookingCameraEnlargementMM);
    WriteRegFloat('Camera Settings', 'Vision Correlation Limit', VisionCorrelationLimit);

    WriteRegFloat('Nozzle Offset Settings', 'Nozzle Offset X', NozzleOffsetX);
    WriteRegFloat('Nozzle Offset Settings', 'Nozzle Offset Y', NozzleOffsetY);
    WriteRegInteger('Nozzle Offset Settings', 'Nozzle Number', NozzleOffsetNozzle);

    WriteRegFloat('Nozzle Offset Settings', 'Nozzle Calibration X', NozzleCalibrateOffsetX);
    WriteRegFloat('Nozzle Offset Settings', 'Nozzle Calibration Y', NozzleCalibrateOffsetY);
    WriteRegFloat('Nozzle Offset Settings', 'Nozzle Calibration Z', NozzleCalibrateOffsetZ);

    WriteRegFloat('Nozzle Changer Settings', 'Nozzle Horizontal', NozzleHorizontal);
    WriteRegFloat('Nozzle Changer Settings', 'Nozzle Up', NozzleUp);
    WriteRegFloat('Nozzle Changer Settings', 'Nozzle Down', NozzleDown);
    WriteRegFloat('Nozzle Changer Settings', 'Nozzle Pick Down', NozzlePickDown);
    WriteRegInteger('Nozzle Changer Settings', 'Nozzle Changer Speed', NozzleChangerSpeed);

    WriteRegFloat('Nozzle Changer Settings', 'Nozzle 1 X', Nozzle1X);
    WriteRegFloat('Nozzle Changer Settings', 'Nozzle 2 X', Nozzle2X);
    WriteRegFloat('Nozzle Changer Settings', 'Nozzle 3 X', Nozzle3X);
    WriteRegFloat('Nozzle Changer Settings', 'Nozzle 4 X', Nozzle4X);
    WriteRegFloat('Nozzle Changer Settings', 'Nozzle 5 X', Nozzle5X);
    WriteRegFloat('Nozzle Changer Settings', 'Nozzle 6 X', Nozzle6X);
    WriteRegFloat('Nozzle Changer Settings', 'Nozzle 7 X', Nozzle7X);
    WriteRegFloat('Nozzle Changer Settings', 'Nozzle 8 X', Nozzle8X);

    WriteRegFloat('Nozzle Changer Settings', 'Nozzle 1 Y', Nozzle1Y);
    WriteRegFloat('Nozzle Changer Settings', 'Nozzle 2 Y', Nozzle2Y);
    WriteRegFloat('Nozzle Changer Settings', 'Nozzle 3 Y', Nozzle3Y);
    WriteRegFloat('Nozzle Changer Settings', 'Nozzle 4 Y', Nozzle4Y);
    WriteRegFloat('Nozzle Changer Settings', 'Nozzle 5 Y', Nozzle5Y);
    WriteRegFloat('Nozzle Changer Settings', 'Nozzle 6 Y', Nozzle6Y);
    WriteRegFloat('Nozzle Changer Settings', 'Nozzle 7 Y', Nozzle7Y);
    WriteRegFloat('Nozzle Changer Settings', 'Nozzle 8 Y', Nozzle8Y);

    WriteRegBoolean('Nozzle Changer Settings', 'Nozzle Changer Enabled', NozzleChangerEnabled);
    WriteRegInteger('Nozzle Changer Settings', 'Nozzle Changer Location', NozzleChangerLocation);

    WriteRegInteger('Current Status', 'Current Nozzle', CurrentNozzle);
    WriteRegBoolean('Current Status', 'Vision Test Mode', VisionTestMode);

    WriteRegString('Current Status', 'Last Project Path', LastProjectPath);
    WriteRegString('Current Status', 'Last Project Name', LastProjectName);

  except
  end;
  FRegIni.Free;
end;

Procedure Save_Status; { save status related settings }
begin
  FRegIni := TRegIniFile.Create('Software\Dockerty Electronics\Pick And Place 2');
  try

    WriteRegInteger('Current Status', 'Current Nozzle', CurrentNozzle);
    WriteRegBoolean('Current Status', 'Vision Test Mode', VisionTestMode);

  except
  end;
  FRegIni.Free;
end;

function Load_Project(Path, Name: string): boolean; { load project }
var
  X: integer;

  Function ReadIniString(Section, Entry, Default: String): String;
  begin
    Result := FIni.ReadString(Section, Entry, Default);
  end;

  Function ReadIniInteger(Section, Entry: String; Default: longint): longint;
  begin
    Result := FIni.ReadInteger(Section, Entry, Default);
  end;

  Function ReadIniBoolean(Section, Entry: String; Default: boolean): boolean;
  begin
    Result := FIni.ReadBool(Section, Entry, Default);
  end;

  Function ReadIniFloat(Section, Entry: String; Default: double): double;
  var
    S: String;
  begin
    S := FIni.ReadString(Section, Entry, '');
    if S <> '' then
    begin
      Try
        Result := StrToFloat(S);
      except
        Result := Default;
      end;
    end
    else
      Result := Default;
  end;

begin
  FIni := TMemIniFile.Create(Path + Name + Project_Extension);
  try

    PanelOriginX := ReadIniFloat('Board Settings', 'Panel Origin X', 0);
    PanelOriginY := ReadIniFloat('Board Settings', 'Panel Origin Y', 0);
    PanelSpacingX := ReadIniFloat('Board Settings', 'Panel Spacing X', 0);
    PanelSpacingY := ReadIniFloat('Board Settings', 'Panel Spacing Y', 0);
    PanelCountX := ReadIniInteger('Board Settings', 'Panel Count X', 1);
    PanelCountY := ReadIniInteger('Board Settings', 'Panel Count Y', 1);
    PanelBoardCount := 1; // auto reset

    BoardHeight := ReadIniFloat('Board Settings', 'Board Height', 0);

    BoardFiducial1SnapName := ReadIniString('Board Settings', 'Board Fiducial 1 Snap Name', '');
    BoardFiducial1X := ReadIniFloat('Board Settings', 'Board Fiducial 1 X', 0);
    BoardFiducial1Y := ReadIniFloat('Board Settings', 'Board Fiducial 1 Y', 0);
    BoardFiducial1Z := ReadIniFloat('Board Settings', 'Board Fiducial 1 Z', 0);
    BoardFiducial2SnapName := ReadIniString('Board Settings', 'Board Fiducial 2 Snap Name', '');
    BoardFiducial2X := ReadIniFloat('Board Settings', 'Board Fiducial 2 X', 0);
    BoardFiducial2Y := ReadIniFloat('Board Settings', 'Board Fiducial 2 Y', 0);
    BoardFiducial2Z := ReadIniFloat('Board Settings', 'Board Fiducial 2 Z', 0);
    BoardDualFiducial := ReadIniBoolean('Board Settings', 'Dual Fiducial', false);
    BoardFiducialEnable := ReadIniBoolean('Board Settings', 'Fiducial Enable', True);

    BoardMetric := ReadIniBoolean('Board Settings', 'Board Metric', True);

    // ******************************************************************************

    PartCount := ReadIniInteger('Part List', 'Part Count', 0); // get part count
    SetLength(PartList, PartCount); // size part array
    if PartCount > 0 then
    begin
      for X := 0 to Pred(PartCount) do
      begin
        PartList[X].Designator := ReadIniString('Part List', 'Part ' + IntToStr(Succ(X)) + ' Designator', '');
        PartList[X].Comment := ReadIniString('Part List', 'Part ' + IntToStr(Succ(X)) + ' Comment', '');
        PartList[X].X := ReadIniFloat('Part List', 'Part ' + IntToStr(Succ(X)) + ' X', 0);
        PartList[X].Y := ReadIniFloat('Part List', 'Part ' + IntToStr(Succ(X)) + ' Y', 0);
        PartList[X].Angle := ReadIniFloat('Part List', 'Part ' + IntToStr(Succ(X)) + ' Angle', 0);
        PartList[X].Feeder := ReadIniInteger('Part List', 'Part ' + IntToStr(Succ(X)) + ' Feeder', 0);
        PartList[X].Enabled := ReadIniBoolean('Part List', 'Part ' + IntToStr(Succ(X)) + ' Enabled', True);
      end;
      CurrentPart := 1; // select part 1
    end
    else
      CurrentPart := 0; // no part selected

    // ******************************************************************************

    FeederCount := ReadIniInteger('Feeder List', 'Feeder Count', 0); // get feeder count
    SetLength(FeederList, FeederCount); // size feeder array
    if FeederCount > 0 then
    begin
      for X := 0 to Pred(FeederCount) do
      begin
        FeederList[X].Description := ReadIniString('Feeder List', 'Feeder ' + IntToStr(Succ(X)) + ' Description', '');
        FeederList[X].SnapName := ReadIniString('Feeder List', 'Feeder ' + IntToStr(Succ(X)) + ' Snap Name', '');
        FeederList[X].Kind := ReadIniInteger('Feeder List', 'Feeder ' + IntToStr(Succ(X)) + ' Kind', 0);
        FeederList[X].Location := ReadIniInteger('Feeder List', 'Feeder ' + IntToStr(Succ(X)) + ' Location', 1);
        FeederList[X].X := ReadIniFloat('Feeder List', 'Feeder ' + IntToStr(Succ(X)) + ' X', 0);
        FeederList[X].Y := ReadIniFloat('Feeder List', 'Feeder ' + IntToStr(Succ(X)) + ' Y', 0);
        FeederList[X].Z := ReadIniFloat('Feeder List', 'Feeder ' + IntToStr(Succ(X)) + ' Z', 0);
        FeederList[X].PartHeight := ReadIniFloat('Feeder List', 'Feeder ' + IntToStr(Succ(X)) + ' Height', 0);
        FeederList[X].Nozzle := ReadIniInteger('Feeder List', 'Feeder ' + IntToStr(Succ(X)) + ' Nozzle', 1);
        FeederList[X].TraySpacingX := ReadIniFloat('Feeder List', 'Feeder ' + IntToStr(Succ(X)) + ' Tray X Spacing', 0);
        FeederList[X].TraySpacingY := ReadIniFloat('Feeder List', 'Feeder ' + IntToStr(Succ(X)) + ' Tray Y Spacing', 0);
        FeederList[X].TrayCountX := ReadIniInteger('Feeder List', 'Feeder ' + IntToStr(Succ(X)) + ' Tray X Count', 1);
        FeederList[X].TrayCountY := ReadIniInteger('Feeder List', 'Feeder ' + IntToStr(Succ(X)) + ' Tray Y Count', 1);
        FeederList[X].TrayPickCount := ReadIniInteger('Feeder List', 'Feeder ' + IntToStr(Succ(X)) + ' Tray Pick Count', 1);
        FeederList[X].PlaceSpeedType := ReadIniInteger('Feeder List', 'Feeder ' + IntToStr(Succ(X)) + ' Place Speed Type', 0);
        FeederList[X].RejectTray := ReadIniInteger('Feeder List', 'Feeder ' + IntToStr(Succ(X)) + ' Reject Tray', 1);
        FeederList[X].Use_Vision := ReadIniBoolean('Feeder List', 'Feeder ' + IntToStr(Succ(X)) + ' Use Vision', True);

        // ************ automatically update location from type **************
        if FeederList[X].Kind = 1 then // tray
        begin
          FeederList[X].Location := 1; // front
        end
        else if FeederList[X].Kind = 2 then // vibration
        begin
          FeederList[X].Location := 0; // right
        end;

      end;
      CurrentFeeder := 1; // select feeder 1
    end
    else
      CurrentFeeder := 0; // no feeder selected

    FiducialX := 0; // reset fiducials
    FiducialY := 0;

    Result := True; // success
  except
    Result := false; // failure
  end;
  FIni.Free;

  // ******************************************************************************

  TrayRefillRequest := 0; // cancel any pending tray refill request

  if Result then
  begin
    ProjectPath := Path; // save project details
    ProjectName := Name;
    LastProjectPath := Path;
    LastProjectName := Name;
  end;
end;

function Save_Project(Path, Name: string): boolean; { save project }
var
  X: integer;

  procedure WriteIniSection(Section: String);
  begin
    Writeln(Tfile, '[' + Section + ']');
  end;

  procedure WriteIniString(Entry, Str: String);
  begin
    Writeln(Tfile, Entry + '=' + Str);
  end;

  procedure WriteIniInteger(Entry: String; Int: integer);
  begin
    Writeln(Tfile, Entry + '=' + IntToStr(Int));
  end;

  procedure WriteIniBoolean(Entry: String; B: boolean);
  begin
    if B then
      Writeln(Tfile, Entry + '=1')
    else
      Writeln(Tfile, Entry + '=0');
  end;

  procedure WriteIniFloat(Entry: String; R: double);
  Var
    Str: String;
  begin
    Str := FloatToStrF(R, ffFixed, 10, 10);
    WriteIniString(Entry, Str);
  end;

begin
  if FileExists(Path + Name + Project_Extension) then // if the file exists
  begin
    if FileExists(Path + Name + '.BAK') then // if the BAK file exists
      try
        DeleteFile(Path + Name + '.BAK'); // delete the old BAK file
      except
      end;

    try // rename the existing file to BAK
      RenameFile(Path + Name + Project_Extension, Path + Name + '.BAK');
    except
    end;
  end;
{$I-}
  IoResult; { flush errors }
  AssignFile(Tfile, Path + Name + Project_Extension);
  Rewrite(Tfile);

  try
    WriteIniSection('Board Settings');

    WriteIniFloat('Panel Origin X', PanelOriginX);
    WriteIniFloat('Panel Origin Y', PanelOriginY);
    WriteIniFloat('Panel Spacing X', PanelSpacingX);
    WriteIniFloat('Panel Spacing Y', PanelSpacingY);
    WriteIniInteger('Panel Count X', PanelCountX);
    WriteIniInteger('Panel Count Y', PanelCountY);

    WriteIniFloat('Board Height', BoardHeight);
    WriteIniString('Board Fiducial 1 Snap Name', BoardFiducial1SnapName);
    WriteIniFloat('Board Fiducial 1 X', BoardFiducial1X);
    WriteIniFloat('Board Fiducial 1 Y', BoardFiducial1Y);
    WriteIniFloat('Board Fiducial 1 Z', BoardFiducial1Z);
    WriteIniString('Board Fiducial 2 Snap Name', BoardFiducial2SnapName);
    WriteIniFloat('Board Fiducial 2 X', BoardFiducial2X);
    WriteIniFloat('Board Fiducial 2 Y', BoardFiducial2Y);
    WriteIniFloat('Board Fiducial 2 Z', BoardFiducial2Z);
    WriteIniBoolean('Dual Fiducial', BoardDualFiducial);
    WriteIniBoolean('Fiducial Enable', BoardFiducialEnable);
    WriteIniBoolean('Board Metric', BoardMetric);

    // ******************************************************************************

    WriteIniSection('Part List');
    WriteIniInteger('Part Count', PartCount); // save part count

    if PartCount > 0 then
      for X := 0 to Pred(PartCount) do
      begin
        WriteIniString('Part ' + IntToStr(Succ(X)) + ' Designator', PartList[X].Designator);
        WriteIniString('Part ' + IntToStr(Succ(X)) + ' Comment', PartList[X].Comment);
        WriteIniFloat('Part ' + IntToStr(Succ(X)) + ' X', PartList[X].X);
        WriteIniFloat('Part ' + IntToStr(Succ(X)) + ' Y', PartList[X].Y);
        WriteIniFloat('Part ' + IntToStr(Succ(X)) + ' Angle', PartList[X].Angle);
        WriteIniInteger('Part ' + IntToStr(Succ(X)) + ' Feeder', PartList[X].Feeder);
        WriteIniBoolean('Part ' + IntToStr(Succ(X)) + ' Enabled', PartList[X].Enabled);
      end;

    // ******************************************************************************

    WriteIniSection('Feeder List');
    WriteIniInteger('Feeder Count', FeederCount); // save feeder count

    if FeederCount > 0 then
      for X := 0 to Pred(FeederCount) do
      begin
        WriteIniString('Feeder ' + IntToStr(Succ(X)) + ' Description', FeederList[X].Description);
        WriteIniString('Feeder ' + IntToStr(Succ(X)) + ' Snap Name', FeederList[X].SnapName);
        WriteIniInteger('Feeder ' + IntToStr(Succ(X)) + ' Kind', FeederList[X].Kind);
        WriteIniInteger('Feeder ' + IntToStr(Succ(X)) + ' Location', FeederList[X].Location);
        WriteIniFloat('Feeder ' + IntToStr(Succ(X)) + ' X', FeederList[X].X);
        WriteIniFloat('Feeder ' + IntToStr(Succ(X)) + ' Y', FeederList[X].Y);
        WriteIniFloat('Feeder ' + IntToStr(Succ(X)) + ' Z', FeederList[X].Z);
        WriteIniFloat('Feeder ' + IntToStr(Succ(X)) + ' Height', FeederList[X].PartHeight);
        WriteIniInteger('Feeder ' + IntToStr(Succ(X)) + ' Nozzle', FeederList[X].Nozzle);
        WriteIniFloat('Feeder ' + IntToStr(Succ(X)) + ' Tray X Spacing', FeederList[X].TraySpacingX);
        WriteIniFloat('Feeder ' + IntToStr(Succ(X)) + ' Tray Y Spacing', FeederList[X].TraySpacingY);
        WriteIniInteger('Feeder ' + IntToStr(Succ(X)) + ' Tray X Count', FeederList[X].TrayCountX);
        WriteIniInteger('Feeder ' + IntToStr(Succ(X)) + ' Tray Y Count', FeederList[X].TrayCountY);
        WriteIniInteger('Feeder ' + IntToStr(Succ(X)) + ' Tray Pick Count', FeederList[X].TrayPickCount);
        WriteIniInteger('Feeder ' + IntToStr(Succ(X)) + ' Place Speed Type', FeederList[X].PlaceSpeedType);
        WriteIniInteger('Feeder ' + IntToStr(Succ(X)) + ' Reject Tray', FeederList[X].RejectTray);
        WriteIniBoolean('Feeder ' + IntToStr(Succ(X)) + ' Use Vision', FeederList[X].Use_Vision);
      end;
  except
  end;

  CloseFile(Tfile);
  Result := (IoResult = 0); { flush errors }
{$I-}
  // ******************************************************************************

  if Result then
  begin
    ProjectPath := Path; // save project details
    ProjectName := Name;
    LastProjectPath := Path;
    LastProjectName := Name;
  end;
end;

function Create_New_Project(Path, Name: string): boolean; { create project }

  procedure WriteIniSection(Section: String);
  begin
    Writeln(Tfile, '[' + Section + ']');
  end;

  procedure WriteIniString(Entry, Str: String);
  begin
    Writeln(Tfile, Entry + '=' + Str);
  end;

  procedure WriteIniInteger(Entry: String; Int: integer);
  begin
    Writeln(Tfile, Entry + '=' + IntToStr(Int));
  end;

  procedure WriteIniBoolean(Entry: String; B: boolean);
  begin
    if B then
      Writeln(Tfile, Entry + '=1')
    else
      Writeln(Tfile, Entry + '=0');
  end;

  procedure WriteIniFloat(Entry: String; R: double);
  Var
    Str: String;
  begin
    Str := FloatToStrF(R, ffFixed, 10, 10);
    WriteIniString(Entry, Str);
  end;

begin
  if FileExists(Path + Name + Project_Extension) then // if the file exists
  begin
    if FileExists(Path + Name + '.BAK') then // if the BAK file exists
      try
        DeleteFile(Path + Name + '.BAK'); // delete the old BAK file
      except
      end;

    try // rename the existing file to BAK
      RenameFile(Path + Name + Project_Extension, Path + Name + '.BAK');
    except
    end;
  end;
{$I-}
  IoResult; { flush errors }
  AssignFile(Tfile, Path + Name + Project_Extension);
  Rewrite(Tfile);

  try
    WriteIniSection('Board Settings');

    WriteIniSection('Part List');
    WriteIniInteger('Part Count', 0); // save part count

    WriteIniSection('Feeder List');
    WriteIniInteger('Feeder Count', 0); // save feeder count

  except
  end;

  CloseFile(Tfile);
  Result := (IoResult = 0); { flush errors }
{$I-}
  // ******************************************************************************

  if Result then
  begin
    ProjectPath := Path; // save project details
    ProjectName := Name;
    LastProjectPath := Path;
    LastProjectName := Name;
  end;
end;

function Load_Feeder(Path, Name: string; var Feeder: FeederRec): boolean;
{ load feeder }
  Function ReadIniString(Section, Entry, Default: String): String;
  begin
    Result := FIni.ReadString(Section, Entry, Default);
  end;

  Function ReadIniInteger(Section, Entry: String; Default: longint): longint;
  begin
    Result := FIni.ReadInteger(Section, Entry, Default);
  end;

  Function ReadIniBoolean(Section, Entry: String; Default: boolean): boolean;
  begin
    Result := FIni.ReadBool(Section, Entry, Default);
  end;

  Function ReadIniFloat(Section, Entry: String; Default: double): double;
  var
    S: String;
  begin
    S := FIni.ReadString(Section, Entry, '');
    if S <> '' then
    begin
      Try
        Result := StrToFloat(S);
      except
        Result := Default;
      end;
    end
    else
      Result := Default;
  end;

begin
  FIni := TMemIniFile.Create(Path + Name + Feeder_Extension);
  try
    Feeder.Description := ReadIniString('Feeder List', 'Feeder Description', '');
    Feeder.SnapName := ReadIniString('Feeder List', 'Feeder Snap Name', '');
    Feeder.Kind := ReadIniInteger('Feeder List', 'Feeder Kind', 0);
    Feeder.Location := ReadIniInteger('Feeder List', 'Feeder Location', 1);
    Feeder.X := ReadIniFloat('Feeder List', 'Feeder X', 0);
    Feeder.Y := ReadIniFloat('Feeder List', 'Feeder Y', 0);
    Feeder.Z := ReadIniFloat('Feeder List', 'Feeder Z', 0);
    Feeder.PartHeight := ReadIniFloat('Feeder List', 'Feeder Height', 0);
    Feeder.Nozzle := ReadIniInteger('Feeder List', 'Feeder Nozzle', 1);
    Feeder.TraySpacingX := ReadIniFloat('Feeder List', 'Feeder Tray X Spacing', 0);
    Feeder.TraySpacingY := ReadIniFloat('Feeder List', 'Feeder Tray Y Spacing', 0);
    Feeder.TrayCountX := ReadIniInteger('Feeder List', 'Feeder Tray X Count', 1);
    Feeder.TrayCountY := ReadIniInteger('Feeder List', 'Feeder Tray Y Count', 1);
    Feeder.TrayPickCount := 1; // reset pick count
    Feeder.PlaceSpeedType := ReadIniInteger('Feeder List', 'Feeder Place Speed Type', 0);
    Feeder.RejectTray := ReadIniInteger('Feeder List', 'Feeder Reject Tray', 1);
    Feeder.Use_Vision := ReadIniBoolean('Feeder List', 'Feeder Use Vision', True);

    // ************ automatically update location from type **************
    if Feeder.Kind = 1 then // tray
    begin
      Feeder.Location := 1; // front
    end
    else if Feeder.Kind = 2 then // vibration
    begin
      Feeder.Location := 0; // right
    end;

    Result := True; // success
  except
    Result := false; // failure
  end;
  FIni.Free;
end;

function Save_Feeder(Path, Name: string; const Feeder: FeederRec): boolean;
{ save feeder }
  procedure WriteIniSection(Section: String);
  begin
    Writeln(Tfile, '[' + Section + ']');
  end;

  procedure WriteIniString(Entry, Str: String);
  begin
    Writeln(Tfile, Entry + '=' + Str);
  end;

  procedure WriteIniInteger(Entry: String; Int: integer);
  begin
    Writeln(Tfile, Entry + '=' + IntToStr(Int));
  end;

  procedure WriteIniBoolean(Entry: String; B: boolean);
  begin
    if B then
      Writeln(Tfile, Entry + '=1')
    else
      Writeln(Tfile, Entry + '=0');
  end;

  procedure WriteIniFloat(Entry: String; R: double);
  Var
    Str: String;
  begin
    Str := FloatToStrF(R, ffFixed, 10, 10);
    WriteIniString(Entry, Str);
  end;

begin
{$I-}
  IoResult; { flush errors }
  AssignFile(Tfile, Path + Name + Feeder_Extension);
  Rewrite(Tfile);

  // ******************************************************************************

  try
    WriteIniSection('Feeder List');

    WriteIniString('Feeder Description', Feeder.Description);
    WriteIniString('Feeder Snap Name', Feeder.SnapName);
    WriteIniInteger('Feeder Kind', Feeder.Kind);
    WriteIniInteger('Feeder Location', Feeder.Location);
    WriteIniFloat('Feeder X', Feeder.X);
    WriteIniFloat('Feeder Y', Feeder.Y);
    WriteIniFloat('Feeder Z', Feeder.Z);
    WriteIniFloat('Feeder Height', Feeder.PartHeight);
    WriteIniInteger('Feeder Nozzle', Feeder.Nozzle);
    WriteIniFloat('Feeder Tray X Spacing', Feeder.TraySpacingX);
    WriteIniFloat('Feeder Tray Y Spacing', Feeder.TraySpacingY);
    WriteIniInteger('Feeder Tray X Count', Feeder.TrayCountX);
    WriteIniInteger('Feeder Tray Y Count', Feeder.TrayCountY);
    WriteIniInteger('Feeder Place Speed Type', Feeder.PlaceSpeedType);
    WriteIniInteger('Feeder Reject Tray', Feeder.RejectTray);
    WriteIniBoolean('Feeder Use Vision', Feeder.Use_Vision);
  except
  end;

  // ******************************************************************************

  CloseFile(Tfile);
  Result := (IoResult = 0); { flush errors }
{$I-}
end;

end.
